
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include "list.h"

struct entry_struct {
	void *data;
	entry next, prev;
};

struct list_struct {
	entry head, tail;
	unsigned int size;
};

void lst_init(list *l) {
	*l = malloc(sizeof(struct list_struct));
	(*l)->size = 0;
	(*l)->head = (*l)->tail = NULL;
}

void lst_free(list *l) {
	lst_clear(*l);
	free(*l);
}

void lst_free_data(list *l, void (*f)(void *)) {
	lst_clear_data(*l, f);
	free(*l);
}

const entry lst_push(list l, void *d) {
	entry e = malloc(sizeof(struct entry_struct));
	e->next = NULL;
	e->prev = l->tail;
	e->data = d;
	if (l->size++ > 0) {
		l->tail->next = e;
		l->tail = e;
	} else {
		l->head = l->tail = e;
	}
	return e;
}

void * lst_pop(list l) {
	entry e = l->head;
	void *r = e->data;
	if (--l->size) {
		l->head = e->next;
		l->head->prev = NULL;
	} else {
		l->head = l->tail = NULL;
	}
	free(e);
	return r;
}

unsigned int lst_size(const list l) {
	return l->size;
}

void lst_clear_data(list l, void (*f)(void *)) {
	entry e = l->head;
	l->head = NULL;
	l->tail = NULL;
	l->size = 0;
	if (f) {
		while (e) {
			entry en = e->next;
			f(e->data);
			free(e);
			e = en;
		}
	} else {
		while (e) {
			entry en = e->next;
			free(en);
			e = en;
		}
	}
}

void lst_clear(list l) {
	entry e = l->head;
	l->head = l->tail = NULL;
	l->size = 0;
	while (e) {
		entry en = e->next;
		free(e);
		e = en;
	}
}

const entry lst_first(const list l) {
	return l->head;
}

const entry lst_last(const list l) {
	return l->tail;
}

const entry lst_next(const entry e) {
	return e->next;
}

const entry lst_prev(const entry e) {
	return e->prev;
}

bool lst_empty(list l) {
	return !(l->size);
}

void * lst_data(const entry e) {
	return e->data;
}

void * lst_remove(list l, const entry e) {
	if (e->prev == NULL) {
		l->head = e->next;
	} else {
		e->prev->next = e->next;
	}
	
	if (e->next == NULL) {
		l->tail = e->prev;
	} else {
		e->next->prev = e->prev;
	}
	
	l->size--;
	void *r = e->data;
	free(e);
	return r;
}

const entry lst_insert(list l, const entry o, void * d, const bool after) {
	entry e = (entry)malloc(sizeof(struct entry_struct));
	e->data = d;
	
	if (!l->size) {
		e->prev = e->next = NULL;
		return l->head = l->tail = e;
	}
	
	if (after) {
		e->prev = o;
		e->next = o->next;
		if (o == l->tail) {
			l->tail = e;
		} else {
			o->next->prev = e;
		}
		o->next = e;
	} else {
		e->next = o;
		e->prev = o->prev;
		if (o == l->head) {
			l->head = e;
		} else {
			o->prev->next = e;
		}
		o->prev = e;
	}
	return e;
}
/*
int main() {
	list l;
	lst_init(&l);
	
	
	printf("insert: %08x\n", lst_push(l, (void *)6));
	printf("insert: %08x\n", lst_push(l, (void *)7));
 	printf("%d\n", lst_pop(l));
	printf("insert: %08x\n", lst_push(l, (void *)8));
	
	printf("head %08x, tail %08x\n", l->head, l->tail);
	entry e;
	lst_forall(e, l)
 		printf("%08x, data: %d\n", e, lst_remove(l, e));
 	
 	while (lst_size(l))
 		printf("%d\n", lst_pop(l));
 	
	lst_free(&l);
	return 0;
}*/
