
#include <stdlib.h>

#include "sequence.h"

struct entry_struct {
	void *data;
	entry next;
};

struct list_struct {
	entry head;
};

void lst_init(list *l) {
	*l = malloc(sizeof(struct list_struct));
	(*l)->head = NULL;
}

void lst_free(list *l) {
	lst_clear(*l);
	free(*l);
}

void lst_free_data(list *l, void (*f)(void *)) {
	lst_clear_data(*l, f);
	free(*l);
}

const entry lst_insert_after(const list l, const entry p, void *d) {
	entry e = malloc(sizeof(struct entry_struct));
	e->data = d;
	e->next = p->next;
	p->next = e;
	return e;
}

const entry lst_push(list l, void *d) {
	entry e = malloc(sizeof(struct entry_struct));
	e->data = d;
	e->next = l->head;
	l->head = e;
	return e;
}

void * lst_pop(list l) {
	entry e = l->head;
	l->head = e->next;
	void *r = e->data;
	free(e);
	return r;
}

void lst_clear_data(list l, void (*f)(void *)) {
	entry e = l->head, en;
	l->head = NULL;
	if (f) {
		while (e) {
			en = e->next;
			f(e->data);
			free(e);
			e = en;
		}
	} else {
		while (e) {
			en = e->next;
			if (e->data)
				free(e->data);
			free(e);
			e = en;
		}
	}
}

void lst_clear(list l) {
	entry e = l->head, en;
	while (e) {
		en = e->next;
		free(e);
		e = en;
	}
}

const entry lst_first(const list l) {
	return l->head;
}

const entry lst_next(const entry e) {
	return e->next;
}

bool lst_empty(const list l) {
	return l->head == NULL;
}

void * lst_data(const entry e) {
	return e->data;
}

/*
#include <stdio.h>

int main() {
	list l;
	lst_init(&l);
	
	
	printf("insert: %08x\n", lst_push(l, (void *)6));
	printf("insert: %08x\n", lst_push(l, (void *)7));
 	printf("%d\n", lst_pop(l));
	printf("insert: %08x\n", lst_push(l, (void *)8));
	
	printf("head %08x\n", l->head);
	entry e;
	
	while (!lst_empty(l))
 		printf("%d\n", lst_pop(l));
 	
	lst_free(&l);
	return 0;
}
*/
